"""
Patch OS when required. For example, allow access to subnet for Windows
and later for Mac (TBD)
"""

import os
import subprocess
from click import echo, secho
from stakkr import docker_actions


def start(project_name: str):
    """For Windows we need to do a few manipulations to be able to access
    our dev environment from the host ...

    Found here: https://github.com/docker/for-win/issues/221#issuecomment-270576243"""

    if os.name not in ['nt']:
        return

    docker_actions.get_client().containers.run(
        'justincormack/nsenter1', remove=True, tty=True, privileged=True, network_mode='none',
        pid_mode='host', command='bin/sh -c "iptables -A FORWARD -j ACCEPT"')

    subnet = docker_actions.get_subnet(project_name)
    switch_ip = docker_actions.get_switch_ip()
    msg = 'We need to create a route for the network {} via {} to work...'
    secho(msg.format(subnet, switch_ip), fg='yellow', nl=False)
    subprocess.call(['route', 'add', subnet, 'MASK', '255.255.255.0', switch_ip])
    echo()


def stop(project_name: str):
    """Opposite actions than start() : cleanup"""

    if os.name not in ['nt']:
        return

    subnet = docker_actions.get_subnet(project_name)
    msg = "Let's remove the route that has been added for {}...".format(subnet)
    secho(msg, fg='yellow', nl=False)
    subprocess.call(['route', 'delete', subnet])
    echo()



"""
Manage DNS Start / Stop by creating the container or attaching to the
current stakkr network
"""

import sys
import click
from docker.errors import DockerException
from stakkr import docker_actions
CT_NAME = 'docker_dns'


def manage_dns(project_name: str, action: str):
    """Control the start or stop of DNS forwarder"""

    dns_started = docker_actions.container_running(CT_NAME)
    if dns_started is True and action == 'stop':
        dns = docker_actions.get_client().containers.get('docker_dns')
        return dns.stop()

    elif action == 'start':
        run_dns(project_name)
        return

    click.echo(click.style('[ERROR]', fg='red') + " DNS already stopped")
    sys.exit(1)


def run_dns(project_name: str):
    """Either add the DNS to the network if started or start it"""

    network = project_name + '_stakkr'

    # Started ? Only attach it to the current network
    if docker_actions.container_running(CT_NAME) is True:
        docker_actions.add_container_to_network(CT_NAME, network)
        return

    # Not started ? Create it and add it to the network
    try:
        volumes = ['/var/run/docker.sock:/tmp/docker.sock', '/etc/resolv.conf:/tmp/resolv.conf']
        docker_actions.get_client().containers.run(
            'mgood/resolvable', remove=True, detach=True, network=network,
            hostname='docker-dns', name=CT_NAME, volumes=volumes)
    except DockerException as error:
        click.echo(click.style('[ERROR]', fg='red') + " Can't start the DNS ...")
        click.echo('Error was -> {}'.format(error))
        sys.exit(1)



"""Gives useful information about the current virtualenv, files locations if
stakkr is installed as a package or directly cloned"""

import os
import sys
from distutils.sysconfig import get_config_vars, get_python_lib


def get_venv_basedir():
    """Returns the base directory of the virtualenv, useful to read configuration and plugins"""

    exec_prefix = get_config_vars()['exec_prefix']
    has_real_prefix = hasattr(sys, 'real_prefix') or (hasattr(sys, 'base_prefix') and sys.base_prefix != sys.prefix)
    if has_real_prefix is False or (hasattr(sys, 'real_prefix') and exec_prefix.startswith(sys.real_prefix)):
        raise EnvironmentError('You must be in a virtual environment')

    return os.path.abspath(get_config_vars()['exec_prefix'] + '/../')


def get_dir(dirname: str):
    """Detects if stakkr is a package or a clone and gives the right path for a directory"""

    staticdir = os.path.dirname(os.path.realpath(__file__)) + '/' + dirname
    if os.path.isdir(staticdir) is True:
        return staticdir

    return get_python_lib() + '/stakkr/' + dirname


def get_file(dirname: str, filename: str):
    """Detects if stakkr is a package or a clone and gives the right path for a file"""

    return get_dir(dirname) + '/' + filename.lstrip('/')



#!/usr/bin/env python
"""
Clean unused containers, images, volumes and networks
That saves a lot of space ... should be removed later by "docker ***** prune
"""

import sys
from subprocess import check_output, CalledProcessError, STDOUT
import click


@click.command(help="""Clean Docker containers, images, volumes and networks
that are not in use""", name="docker-clean")
@click.option('--force', '-f', help="Do it", is_flag=True)
@click.option('--verbose', '-v', help="Display more information about what is removed",
              is_flag=True)
def clean(force: bool, verbose: bool):
    """See command help"""

    click.secho('Clean Docker stopped containers, images, volumes and networks', fg='green')

    remove_containers(force, verbose)
    remove_images(force)
    remove_volumes(force, verbose)
    remove_networks(force, verbose)

    if force is False:
        click.secho("\n--force is not set so I won't do anything", fg='red')


def remove_containers(force: bool, verbose: bool):
    """Remove exited containers"""

    res = _exec_cmd(['docker', 'ps', '--no-trunc', '-a', '-q', '-f', 'status=exited'])
    containers = res.splitlines()
    if len(containers) is 0:
        print('- No exited container to remove')
        return

    click.echo('- Removing {} exited container(s)'.format(len(containers)))

    for container in containers:
        _display_entry_info('container', container.decode(), verbose)
        _remove_entry('container', container.decode(), force)


def remove_images(force: bool):
    """Prune all images"""

    res = _exec_cmd(['docker', 'image', 'ls'])
    images = res.splitlines()
    if len(images) is 1:
        click.echo('- No image to remove')
        return

    click.echo('- Removing {} image(s)'.format(len(images) - 1))
    if force is True:
        _prune_images()


def remove_networks(force: bool, verbose: bool):
    """Remove custom networks (not systems)"""

    res = _exec_cmd(['docker', 'network', 'ls', '--no-trunc', '-q', '--filter', 'type=custom'])
    networks = res.splitlines()
    if len(networks) is 0:
        click.echo('- No network to remove')
        return

    click.echo('- Removing {} exited networks(s)'.format(len(networks)))

    for network in networks:
        _display_entry_info('network', network.decode(), verbose)
        _remove_entry('network', network.decode(), force)


def remove_volumes(force: bool, verbose: bool):
    """Remove dangling volumes"""

    res = _exec_cmd(['docker', 'volume', 'ls', '-q', '-f', 'dangling=true'])
    volumes = res.splitlines()
    if len(volumes) is 0:
        click.echo('- No volume to remove')
        return

    click.echo('- Removing {} exited volumes(s)'.format(len(volumes)))

    for volume in volumes:
        if verbose is True:
            click.echo('  Removing volume {}'.format(volume.decode()))

        _remove_entry('volume', volume.decode(), force)


def _display_entry_info(entry_type: str, entry: str, verbose: bool):
    if verbose is False:
        return

    base_cmd = ['docker'] if entry_type == 'container' else ['docker', entry_type]
    res = _exec_cmd(base_cmd + ['inspect', '--format={{.Name}}', entry])
    info = res.splitlines()[0]
    click.echo('  Removing {} {}'.format(entry_type, info.decode()))


def _exec_cmd(cmd: list):
    try:
        return check_output(cmd, stderr=STDOUT)
    except CalledProcessError:
        click.secho('  Error running command : "{}"'.format(' '.join(cmd)), fg='red')


def _prune_images():
    _exec_cmd(['docker', 'image', 'prune', '--all', '--force'])


def _remove_entry(entry_type: str, entry: str, force: bool):
    if force is False:
        return

    base_cmd = ['docker'] if entry_type == 'container' else ['docker', entry_type]
    _exec_cmd(base_cmd + ['rm', entry])


def main():
    """Main function when the CLI Script is called directly"""

    try:
        clean()
    except Exception as error:
        msg = click.style(r""" ______ _____  _____   ____  _____
|  ____|  __ \|  __ \ / __ \|  __ \
| |__  | |__) | |__) | |  | | |__) |
|  __| |  _  /|  _  /| |  | |  _  /
| |____| | \ \| | \ \| |__| | | \ \
|______|_|  \_\_|  \_\\____/|_|  \_\

""", fg='yellow')
        msg += click.style('{}'.format(error), fg='red')

        click.echo(msg)
        click.echo()
        # raise error
        sys.exit(1)


if __name__ == '__main__':
    main()



"""
CLI Entry Point. From click, build stakkr and give all options to manage
services to be launched, stopped, etc.
"""

import sys
import click
from click_plugins import with_plugins
from pkg_resources import iter_entry_points
from stakkr import dns as dns_manager, package_utils
from stakkr.docker_actions import get_running_containers_name


@with_plugins(iter_entry_points('stakkr.plugins'))
@click.group(help="""Main CLI Tool that easily create / maintain
a stack of services, for example for web development.

Read the configuration file and setup the required services by
linking and managing everything for you.""")
@click.version_option('3.6.2')
@click.option('--config', '-c', help='Change the configuration file')
@click.option('--debug/--no-debug', '-d', default=False)
@click.option('--verbose', '-v', is_flag=True)
@click.pass_context
def stakkr(ctx, config, debug, verbose):
    """click group, set context and main object"""

    from stakkr.actions import StakkrActions

    # Add the virtual env in the path
    venv_base = package_utils.get_venv_basedir()
    sys.path.append(venv_base)

    ctx.obj['CONFIG'] = config
    ctx.obj['DEBUG'] = debug
    ctx.obj['VERBOSE'] = verbose
    ctx.obj['STAKKR'] = StakkrActions(venv_base, ctx.obj)
    ctx.obj['CTS'] = get_running_containers_name(ctx.obj['STAKKR'].project_name)


@stakkr.command(help="""Enter a container to perform direct actions such as
install packages, run commands, etc.""")
@click.option('--user', '-u', help="User's name. Valid choices : www-data or root")
@click.option('--tty/--no-tty', '-t/ ', is_flag=True, default=True, help="Use a TTY")
@click.argument('container', required=True)
@click.pass_context
def console(ctx, container: str, user: str, tty: bool):
    """See command Help"""

    if len(ctx.obj['CTS']) is not 0:
        ct_choice = click.Choice(ctx.obj['CTS'])
        ct_choice.convert(container, None, ctx)

    ctx.obj['STAKKR'].console(container, _get_cmd_user(user, container), tty)


@stakkr.command(help="""Start or Stop the DNS forwarder.
Useful to access your containers directly by their names.
Does not work under Windows as we can't mount /etc/resolv.conf.

Valid values for ACTION : 'start' or 'stop'""", name="dns")
@click.argument('action', required=True, type=click.Choice(['start', 'stop']))
@click.pass_context
def dns(ctx, action: str):
    """See command Help"""

    dns_manager.manage_dns(ctx.obj['STAKKR'].project_name, action)
    action = click.style('[{}]'.format(action.upper()), fg='green')
    click.echo(action + ' DNS forwarder ...', nl=False)
    click.echo('Wait a little before calling services by their DNS')
    _show_status(ctx)


@stakkr.command(help="""Execute a command into a container.

Examples:\n
- ``stakkr -v exec mysql mysqldump -p'$MYSQL_ROOT_PASSWORD' mydb > /tmp/backup.sql``\n
- ``stakkr exec php php -v`` : Execute the php binary in the php container with option -v\n
- ``stakkr exec apache service apache2 restart``\n
""", name='exec', context_settings=dict(ignore_unknown_options=True, allow_interspersed_args=False))
@click.pass_context
@click.option('--user', '-u', help="User's name. Be careful, each container have its own users.")
@click.option('--tty/--no-tty', '-t/ ', is_flag=True, default=True, help="Use a TTY")
@click.argument('container', required=True)
@click.argument('command', required=True, nargs=-1, type=click.UNPROCESSED)
def exec_cmd(ctx, user: str, container: str, command: tuple, tty: bool):
    """See command Help"""

    if len(ctx.obj['CTS']) is not 0:
        click.Choice(ctx.obj['CTS']).convert(container, None, ctx)

    ctx.obj['STAKKR'].exec_cmd(container, _get_cmd_user(user, container), command, tty)


@stakkr.command(help="""`stakkr mysql` is a wrapper for the mysql binary
located in the mysql service.

You can run any mysql command as root, such as :\n
- ``stakkr mysql -e "CREATE DATABASE mydb"`` to create a DB from outside\n
- ``stakkr mysql`` to enter the mysql console\n
- ``cat myfile.sql | stakkr mysql --no-tty mydb`` to import a file from outside to mysql\n

For scripts, you must use the relative path.
""", context_settings=dict(ignore_unknown_options=True, allow_interspersed_args=False))
@click.pass_context
@click.option('--tty/--no-tty', '-t/ ', is_flag=True, default=True, help="Use a TTY")
@click.argument('command', nargs=-1, type=click.UNPROCESSED)
def mysql(ctx, tty: bool, command: tuple):
    """See command Help"""

    command = ('mysql', '-p$MYSQL_ROOT_PASSWORD') + command
    ctx.invoke(exec_cmd, user='root', container='mysql', command=command, tty=tty)


@stakkr.command(help='Required to be launched if you install a new plugin', name="refresh-plugins")
@click.pass_context
def refresh_plugins(ctx):
    """See command Help"""

    from stakkr.plugins import add_plugins

    print(click.style('Adding plugins from plugins/', fg='green'))
    plugins = add_plugins()
    if len(plugins) is 0:
        print(click.style('No plugin to add', fg='yellow'))
        exit(0)

    print()
    print(click.style('Plugins refreshed', fg='green'))


@stakkr.command(help="Restart all containers")
@click.option('--pull', '-p', help="Force a pull of the latest images versions", is_flag=True)
@click.option('--recreate', '-r', help="Recreate all containers", is_flag=True)
@click.pass_context
def restart(ctx, pull: bool, recreate: bool):
    """See command Help"""

    print(click.style('[RESTARTING]', fg='green') + ' your stakkr services')
    try:
        ctx.invoke(stop)
    except Exception:
        print()

    ctx.invoke(start, pull=pull, recreate=recreate)


@stakkr.command(help="Start containers defined in compose.ini")
@click.option('--pull', '-p', help="Force a pull of the latest images versions", is_flag=True)
@click.option('--recreate', '-r', help="Recreate all containers", is_flag=True)
@click.pass_context
def start(ctx, pull: bool, recreate: bool):
    """See command Help"""

    print(click.style('[STARTING]', fg='green') + ' your stakkr services')
    ctx.obj['STAKKR'].start(pull, recreate)
    _show_status(ctx)


@stakkr.command(help="Display a list of running containers")
@click.pass_context
def status(ctx):
    """See command Help"""

    ctx.obj['STAKKR'].status()


@stakkr.command(help="Stop the services")
@click.pass_context
def stop(ctx):
    """See command Help"""

    print(click.style('[STOPPING]', fg='yellow') + ' your stakkr services')
    ctx.obj['STAKKR'].stop()


def _get_cmd_user(user: str, container: str):
    users = {'apache': 'www-data', 'nginx': 'www-data', 'php': 'www-data'}

    cmd_user = 'root' if user is None else user
    if container in users and user is None:
        cmd_user = users[container]

    return cmd_user


def _show_status(ctx):
    services_ports = ctx.obj['STAKKR'].get_services_ports()
    if services_ports == '':
        print('\nServices Status:')
        ctx.invoke(status)
        return

    print('\nServices URLs :')
    print(services_ports)


def debug_mode():
    """Guess if we are in debug mode, useful to display runtime errors"""

    if '--debug' in sys.argv or '-d' in sys.argv:
        return True

    return False


def main():
    """Main function when the CLI Script is called directly"""

    try:
        stakkr(obj={})
    except Exception as error:
        msg = click.style(r""" ______ _____  _____   ____  _____
|  ____|  __ \|  __ \ / __ \|  __ \
| |__  | |__) | |__) | |  | | |__) |
|  __| |  _  /|  _  /| |  | |  _  /
| |____| | \ \| | \ \| |__| | | \ \
|______|_|  \_\_|  \_\\____/|_|  \_\

""", fg='yellow')
        msg += click.style('{}'.format(error), fg='red')
        print(msg + '\n', file=sys.stderr)

        if debug_mode() is True:
            raise error

        sys.exit(1)


if __name__ == '__main__':
    main()



"""Module used by setup.py to find plugins to load with click"""

import re
import subprocess
from os import listdir, path


def add_plugins():
    """Read the plugins directory, get the subfolders from it and look for .py files"""

    if path.isdir('plugins') is False:
        return []

    _remove_plugins()
    folders = _get_subfolders('plugins')

    plugins = []
    for folder in folders:
        plugins = _add_plugin_from_dir(plugins, 'plugins/{}/'.format(folder))

    return sorted(plugins)


def _add_plugin_from_dir(plugins: list, full_path: str):
    files = _get_files_from_folder(full_path)
    if len(files) is 0:
        print('  -> No plugin found in "{}"'.format(full_path))
        return plugins

    plugin_name = full_path.strip('/').split('/')[1]
    try:
        cmd_install = ['pip', 'install', '-e', full_path]
        subprocess.check_call(cmd_install, stdout=subprocess.DEVNULL)
    except Exception as error:
        msg = 'Problem installing {} (Reason: {})'.format(plugin_name[:-3], error)
        raise TypeError(msg)

    print('  -> Plugin "{}" added'.format(plugin_name))
    plugins.append(plugin_name)

    return plugins


def _get_files_from_folder(full_path: str):
    files = listdir(full_path)

    return [filename for filename in files if filename == 'setup.py']


def _get_subfolders(directory: str):
    subfolders = listdir(directory)

    return [folder for folder in subfolders if path.isdir('{}/{}'.format(directory, folder))]


def _remove_plugins():
    cmd = ['pip', 'freeze']
    res = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    regex = re.compile('.*stakkr([0-9a-z]+).*$', re.IGNORECASE)

    for line in res.stdout:
        plugin = re.search(regex, line.decode())
        if plugin is None:
            continue

        plugin_name = 'Stakkr{}'.format(plugin.group(1))
        print('  -> Cleaning "{}"'.format(plugin_name))
        subprocess.check_call(['pip', 'uninstall', '-y', plugin_name], stdout=subprocess.DEVNULL)



"""
Stakkr main controller. Used by the CLI to do all its actions
"""

import os
import subprocess
import sys
import click
from clint.textui import colored, puts, columns
from stakkr import command, docker_actions, os_patch, package_utils
from stakkr.configreader import Config


class StakkrActions():
    """Main class that does actions asked in the cli"""

    _services_to_display = {
        'adminer': {'name': 'Adminer', 'url': 'http://{}'},
        'apache': {'name': 'Web Server', 'url': 'http://{}'},
        'mailcatcher': {'name': 'Mailcatcher (fake SMTP)', 'url': 'http://{}', 'extra_port': 25},
        'maildev': {'name': 'Maildev (Fake SMTP)', 'url': 'http://{}', 'extra_port': 25},
        'nginx': {'name': 'Web Server', 'url': 'http://{}'},
        'portainer': {'name': 'Portainer (Docker GUI)', 'url': 'http://{}'},
        'phpmyadmin': {'name': 'PhpMyAdmin', 'url': 'http://{}'},
        'xhgui': {'name': 'XHGui (PHP Profiling)', 'url': 'http://{}'}
    }


    def __init__(self, base_dir: str, ctx: dict):
        # Work with directories and move to the right place
        self.stakkr_base_dir = base_dir
        self.context = ctx
        self.cwd_abs = os.getcwd()
        self.cwd_relative = self._get_relative_dir()
        os.chdir(self.stakkr_base_dir)

        # Set some general variables
        self.config_file = ctx['CONFIG']
        self.compose_base_cmd = self._get_compose_base_cmd()
        self.cts = []
        self.running_cts = []

        # Get info from config
        self.config = self._get_config()
        self.main_config = self.config['main']
        self.project_name = self.main_config.get('project_name')


    def console(self, container: str, user: str, tty: bool):
        """Enter a container. Stakkr will try to guess the right shell"""

        docker_actions.check_cts_are_running(self.project_name)

        tty = 't' if tty is True else ''
        ct_name = docker_actions.get_ct_name(container)
        cmd = ['docker', 'exec', '-u', user, '-i' + tty]
        cmd += [docker_actions.get_ct_name(container), docker_actions.guess_shell(ct_name)]
        subprocess.call(cmd)

        command.verbose(self.context['VERBOSE'], 'Command : "' + ' '.join(cmd) + '"')


    def get_services_ports(self):
        """Once started, stakkr displays a message with the list of launched containers."""

        dns_started = docker_actions.container_running('docker_dns')

        cts = docker_actions.get_running_containers(self.project_name)[1]

        text = ''
        for ct_id, ct_info in cts.items():
            if ct_info['compose_name'] not in self._services_to_display:
                continue
            options = self._services_to_display[ct_info['compose_name']]
            url = get_url(options['url'], ct_info['compose_name'], dns_started)
            name = colored.yellow(options['name'])
            text += '  - For {}'.format(name).ljust(55, ' ') + ' : ' + url + '\n'

            if len(ct_info['ports']) > 0:
                text += ' '*4 + 'If you are using a Mac, do not use the ip but '
                text += '"http://localhost:PORT" where PORT is '
                text += ' or '.join(ct_info['ports']) + '\n'

            if 'extra_port' in options:
                port = str(options['extra_port'])
                text += ' '*4 + 'In your containers use the host '
                text += '"{}" and port {}\n'.format(ct_info['compose_name'], port)

        return text


    def exec_cmd(self, container: str, user: str, args: tuple, tty: bool):
        """Run a command from outside to any container. Wrapped into /bin/sh"""

        docker_actions.check_cts_are_running(self.project_name)

        # Protect args to avoid strange behavior in exec
        args = ['"{}"'.format(arg) for arg in args]

        tty = 't' if tty is True else ''
        ct_name = docker_actions.get_ct_name(container)
        cmd = ['docker', 'exec', '-u', user, '-i' + tty, ct_name, 'sh', '-c']
        cmd += ["""test -d "/var/{0}" && cd "/var/{0}" ; exec {1}""".format(self.cwd_relative, ' '.join(args))]
        command.verbose(self.context['VERBOSE'], 'Command : "' + ' '.join(cmd) + '"')
        subprocess.call(cmd, stdin=sys.stdin)


    def start(self, pull: bool, recreate: bool):
        """If not started, start the containers defined in config"""

        verb = self.context['VERBOSE']
        debug = self.context['DEBUG']

        try:
            docker_actions.check_cts_are_running(self.project_name)
            puts(colored.yellow('[INFO]') + ' stakkr is already started ...')
            sys.exit(0)
        except SystemError:
            pass

        if pull is True:
            command.launch_cmd_displays_output(self.compose_base_cmd + ['pull'], verb, debug, True)

        recreate_param = '--force-recreate' if recreate is True else '--no-recreate'
        cmd = self.compose_base_cmd + ['up', '-d', recreate_param, '--remove-orphans']
        command.verbose(self.context['VERBOSE'], 'Command: ' + ' '.join(cmd))
        command.launch_cmd_displays_output(cmd, verb, debug, True)

        self.running_cts, self.cts = docker_actions.get_running_containers(self.project_name)
        if self.running_cts is 0:
            raise SystemError("Couldn't start the containers, run the start with '-v' and '-d'")

        os_patch.start(self.project_name)
        self._run_iptables_rules()
        self._run_services_post_scripts()


    def status(self):
        """Returns a nice table with the list of started containers"""

        try:
            docker_actions.check_cts_are_running(self.project_name)
        except SystemError:
            puts(colored.yellow('[INFO]') + ' stakkr is currently stopped')
            sys.exit(0)

        dns_started = docker_actions.container_running('docker_dns')

        self._print_status_headers(dns_started)
        self._print_status_body(dns_started)


    def stop(self):
        """If started, stop the containers defined in config. Else throw an error"""

        verb = self.context['VERBOSE']
        debug = self.context['DEBUG']

        docker_actions.check_cts_are_running(self.project_name)
        command.launch_cmd_displays_output(self.compose_base_cmd + ['stop'], verb, debug, True)
        os_patch.stop(self.project_name)

        self.running_cts, self.cts = docker_actions.get_running_containers(self.project_name)
        if self.running_cts is not 0:
            raise SystemError("Couldn't stop services ...")


    def _call_service_post_script(self, service: str):
        service_script = package_utils.get_file('static', 'services/{}.sh'.format(service))
        if os.path.isfile(service_script) is True:
            cmd = ['bash', service_script, docker_actions.get_ct_item(service, 'name')]
            subprocess.call(cmd)
            command.verbose(self.context['VERBOSE'], 'Service Script : ' + ' '.join(cmd))


    def _get_compose_base_cmd(self):
        if self.context['CONFIG'] is None:
            return ['stakkr-compose']

        return ['stakkr-compose', '-c', self.context['CONFIG']]


    def _get_config(self):
        config = Config(self.config_file)
        main_config = config.read()
        if main_config is False:
            config.display_errors()
            sys.exit(1)

        return main_config


    def _get_relative_dir(self):
        if self.cwd_abs.startswith(self.stakkr_base_dir):
            return self.cwd_abs[len(self.stakkr_base_dir):].lstrip('/')

        return ''


    def _print_status_headers(self, dns_started: bool):
        host_ip = (colored.green('HostName' if dns_started else 'IP'))

        puts(columns(
            [(colored.green('Container')), 16], [host_ip, 25], [(colored.green('Ports')), 25],
            [(colored.green('Image')), 32],
            [(colored.green('Docker ID')), 15], [(colored.green('Docker Name')), 25]
            ))

        puts(columns(
            ['-'*16, 16], ['-'*25, 25], ['-'*25, 25],
            ['-'*32, 32],
            ['-'*15, 15], ['-'*25, 25]
            ))


    def _print_status_body(self, dns_started: bool):
        self.running_cts, self.cts = docker_actions.get_running_containers(self.project_name)

        for container in sorted(self.cts.keys()):
            ct_data = self.cts[container]
            if ct_data['ip'] == '':
                continue

            host_ip = ct_data['name'] if dns_started else ct_data['ip']

            puts(columns(
                [ct_data['compose_name'], 16], [host_ip, 25], [', '.join(ct_data['ports']), 25],
                [ct_data['image'], 32],
                [ct_data['id'][:12], 15], [ct_data['name'], 25]
                ))


    def _run_iptables_rules(self):
        """For some containers we need to add iptables rules added from the config"""

        block_config = self.config['network-block']
        for service, ports in block_config.items():
            error, msg = docker_actions.block_ct_ports(service, ports, self.project_name)
            if error is True:
                click.secho(msg, fg='red')
                continue

            command.verbose(self.context['VERBOSE'], msg)


    def _run_services_post_scripts(self):
        """A service can have a .sh file that will be executed once it's started.
        Useful to override some actions of the classical /run.sh

        """

        if os.name == 'nt':
            click.secho('Could not run service post scripts under Windows', fg='red')
            return

        for service in self.main_config.get('services'):
            self._call_service_post_script(service)


def get_url(service_url: str, service: str, dns_started: bool):
    """Build URL to be displayed"""

    service_name = docker_actions.get_ct_name(service)
    service_ip = docker_actions.get_ct_item(service, 'ip')

    return service_url.format(service_name if dns_started else service_ip)



import os
from distutils.core import setup
from stakkr.setup import StakkrPostInstall


here = os.path.abspath(os.path.dirname(__file__))
extra_packages = []
if os.name == 'nt':
    extra_packages.append('pypiwin32')


__version__ = '3.6.2'

# Get the long description from the README file
with open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='stakkr',
    version=__version__,
    description='A configurable stack based on docker to run any combination of services (PHP, MySQL, Apache, Nginx, Mongo, etc..)',
    long_description=long_description,
    url='http://github.com/edyan/stakkr',
    author='Emmanuel Dyan',
    author_email='emmanueldyan@gmail.com',
    license='Apache 2.0',
    packages=['stakkr'],
    py_modules=['stakkr'],
    entry_points='''[console_scripts]
docker-clean=stakkr.docker_clean:main
stakkr=stakkr.cli:main
stakkr-init=stakkr.setup:init
stakkr-compose=stakkr.stakkr_compose:cli''',
    include_package_data=True,
    install_requires=[
        'click-plugins==1.0.3',
        'clint==0.5.1',
        'configobj==5.0.6',
        # Docker compose includes requests and docker
        'docker-compose<1.20.0'
        ] + extra_packages,
    cmdclass={'install': StakkrPostInstall},
    classifiers=[
        'Development Status :: 4 - Beta',
        # 'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development',
        'Topic :: System :: Distributed Computing',
        'Topic :: System :: Systems Administration',
    ],
    keywords='docker,docker-compose,python,stack,development,lamp,vagrant',
)



"""
Setup post actions, used in the main setup.py
"""

import os
import shutil
import sys
from setuptools.command.install import install
from stakkr import package_utils


try:
    import click

    @click.command(help="""Initialize for the first time stakkr by copying
templates and directory structure""")
    @click.option('--force', '-f', help="Force recreate directories structure",
                  is_flag=True)
    def init(force: bool):
        """CLI Entry point, when initializing stakkr manually"""

        config_file = package_utils.get_venv_basedir() + '/conf/compose.ini'
        if os.path.isfile(config_file) and force is False:
            click.secho('Config file (conf/compose.ini) already present. Leaving.', fg='yellow')
            return

        msg = 'Config file (conf/compose.ini) not present, do not forget to create it'
        click.secho(msg, fg='yellow')
        _post_install(force)
except ImportError:
    def init():
        """If click is not installed, display that message"""

        print('Stakkr has not been installed yet')
        sys.exit(1)


def _post_install(force: bool = False):
    print('Post Installation : create templates')

    venv_dir = package_utils.get_venv_basedir()
    # If already installed don't do anything
    if os.path.isfile(venv_dir + '/conf/compose.ini'):
        return

    required_dirs = [
        'conf/mysql-override',
        'conf/php-fpm-override',
        'conf/xhgui-override',
        'data',
        'home/www-data',
        'home/www-data/bin',
        'logs',
        'plugins',
        'services',
        'www'
    ]
    for required_dir in required_dirs:
        _create_dir(venv_dir, required_dir, force)

    required_tpls = [
        '.env',
        'bash_completion',
        'conf/compose.ini.tpl',
        'conf/mysql-override/mysqld.cnf',
        'conf/php-fpm-override/example.conf',
        'conf/php-fpm-override/README',
        'conf/xhgui-override/config.php',
        'home/www-data/.bashrc'
    ]
    for required_tpl in required_tpls:
        _copy_file(venv_dir, required_tpl, force)


def _create_dir(venv_dir: str, dir_name: str, force: bool):
    dir_name = venv_dir + '/' + dir_name.lstrip('/')
    if os.path.isdir(dir_name) and force is False:
        return

    if not os.path.isdir(dir_name):
        os.makedirs(dir_name)


def _copy_file(venv_dir: str, source_file: str, force: bool):
    full_path = package_utils.get_file('tpls', source_file)
    dest_file = venv_dir + '/' + source_file
    if os.path.isfile(dest_file) and force is False:
        print('  - {} exists, do not overwrite'.format(source_file))
        return

    print('  - {} written'.format(source_file))
    try:
        shutil.copy(full_path, dest_file)
    except Exception:
        msg = "Error trying to copy {} .. check that the file is there ...".format(full_path)
        print(msg, file=sys.stderr)


class StakkrPostInstall(install):
    """Class called by the main setup.py"""

    def __init__(self, *args, **kwargs):
        super(StakkrPostInstall, self).__init__(*args, **kwargs)

        try:
            package_utils.get_venv_basedir()
            _post_install(False)
        except OSError:
            msg = 'You must run setup.py from a virtualenv if you want to have '
            msg += 'the templates installed'
            print(msg)

